@@ -1,9 +1,9 @@
 /*
  * psql - the PostgreSQL interactive terminal
  *
- * Copyright (c) 2000-2003, PostgreSQL Global Development Group
+ * Copyright (c) 2000-2005, PostgreSQL Global Development Group
  *
- * $Header: /cvsroot/pgsql-server/src/bin/psql/describe.c,v 1.87 2003/10/26 02:53:45 momjian Exp $
+ * $PostgreSQL: pgsql/src/bin/psql/describe.c,v 1.111.4.2 2006/10/10 16:15:28 tgl Exp $
  */
 #include "postgres_fe.h"
 #include "describe.h"
@@ -39,31 +39,8 @@ static void processNamePattern(PQExpBuffer buf, const char *pattern,
 				   const char *schemavar, const char *namevar,
 				   const char *altnamevar, const char *visibilityrule);
 
-
-static void *
-xmalloc(size_t size)
-{
-	void	   *tmp;
-
-	tmp = malloc(size);
-	if (!tmp)
-	{
-		psql_error("out of memory\n");
-		exit(EXIT_FAILURE);
-	}
-	return tmp;
-}
-
-static void *
-xmalloczero(size_t size)
-{
-	void	   *tmp;
-
-	tmp = xmalloc(size);
-	memset(tmp, 0, size);
-	return tmp;
-}
-
+static void add_tablespace_footer(char relkind, Oid tablespace,
+					  char **footers, int *count, PQExpBufferData buf);
 
 /*----------------
  * Handlers for various slash commands displaying some sort of list
@@ -125,6 +102,59 @@ describeAggregates(const char *pattern, bool verbose)
 	return true;
 }
 
+/* \db
+ * Takes an optional regexp to select particular tablespaces
+ */
+bool
+describeTablespaces(const char *pattern, bool verbose)
+{
+	PQExpBufferData buf;
+	PGresult   *res;
+	printQueryOpt myopt = pset.popt;
+
+	if (pset.sversion < 80000)
+	{
+		fprintf(stderr, _("The server version (%d) does not support tablespaces.\n"),
+				pset.sversion);
+		return true;
+	}
+
+	initPQExpBuffer(&buf);
+
+	printfPQExpBuffer(&buf,
+					  "SELECT spcname AS \"%s\",\n"
+					"  pg_catalog.pg_get_userbyid(spcowner) AS \"%s\",\n"
+					  "  spclocation AS \"%s\"",
+					  _("Name"), _("Owner"), _("Location"));
+
+	if (verbose)
+		appendPQExpBuffer(&buf,
+						  ",\n  spcacl as \"%s\"",
+						  _("Access privileges"));
+
+	appendPQExpBuffer(&buf,
+					  "\nFROM pg_catalog.pg_tablespace\n");
+
+	processNamePattern(&buf, pattern, false, false,
+					   NULL, "spcname", NULL,
+					   NULL);
+
+	appendPQExpBuffer(&buf, "ORDER BY 1;");
+
+	res = PSQLexec(buf.data, false);
+	termPQExpBuffer(&buf);
+	if (!res)
+		return false;
+
+	myopt.nullPrint = NULL;
+	myopt.title = _("List of tablespaces");
+
+	printQuery(res, &myopt, pset.queryFout);
+
+	PQclear(res);
+	return true;
+}
+
 
 /* \df
  * Takes an optional regexp to select particular functions
@@ -139,12 +169,12 @@ describeFunctions(const char *pattern, bool verbose)
 	initPQExpBuffer(&buf);
 
 	printfPQExpBuffer(&buf,
-			"SELECT CASE WHEN p.proretset THEN 'setof ' ELSE '' END ||\n"
-			  "  pg_catalog.format_type(p.prorettype, NULL) as \"%s\",\n"
-					  "  n.nspname as \"%s\",\n"
+					  "SELECT n.nspname as \"%s\",\n"
 					  "  p.proname as \"%s\",\n"
-				  "  pg_catalog.oidvectortypes(p.proargtypes) as \"%s\"",
-					  _("Result data type"), _("Schema"), _("Name"),
+					  "  CASE WHEN p.proretset THEN 'setof ' ELSE '' END ||\n"
+					  "  pg_catalog.format_type(p.prorettype, NULL) as \"%s\",\n"
+					  "  pg_catalog.oidvectortypes(p.proargtypes) as \"%s\"",
+					  _("Schema"), _("Name"), _("Result data type"), 
 					  _("Argument data types"));
 
 	if (verbose)
@@ -180,7 +210,7 @@ describeFunctions(const char *pattern, bool verbose)
 					   "n.nspname", "p.proname", NULL,
 					   "pg_catalog.pg_function_is_visible(p.oid)");
 
-	appendPQExpBuffer(&buf, "ORDER BY 2, 3, 1, 4;");
+	appendPQExpBuffer(&buf, "ORDER BY 1, 2, 3, 4;");
 
 	res = PSQLexec(buf.data, false);
 	termPQExpBuffer(&buf);
@@ -377,11 +407,12 @@ permissionsList(const char *pattern)
 	printfPQExpBuffer(&buf,
 					  "SELECT n.nspname as \"%s\",\n"
 					  "  c.relname as \"%s\",\n"
+					  "  CASE c.relkind WHEN 'r' THEN '%s' WHEN 'v' THEN '%s' WHEN 'S' THEN '%s' END as \"%s\",\n"
 					  "  c.relacl as \"%s\"\n"
 					  "FROM pg_catalog.pg_class c\n"
 	"     LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace\n"
 					  "WHERE c.relkind IN ('r', 'v', 'S')\n",
-					  _("Schema"), _("Table"), _("Access privileges"));
+					  _("Schema"), _("Name"), _("table"), _("view"), _("sequence"), _("Type"), _("Access privileges"));
 
 	/*
 	 * Unless a schema pattern is specified, we suppress system and temp
@@ -574,6 +605,8 @@ objectDescription(const char *pattern)
  *
  * This routine finds the tables to be displayed, and calls
  * describeOneTableDetails for each one.
+ *
+ * verbose: if true, this is \d+
  */
 bool
 describeTableDetails(const char *pattern, bool verbose)
@@ -660,11 +693,13 @@ describeOneTableDetails(const char *schemaname,
 	int			numrows = 0;
 	struct
 	{
-		bool		hasindex;
-		char		relkind;
 		int16		checks;
 		int16		triggers;
+		char		relkind;
+		bool		hasindex;
 		bool		hasrules;
+		bool		hasoids;
+		Oid			tablespace;
 	}			tableinfo;
 	bool		show_modifiers = false;
 	bool		retval;
@@ -677,8 +712,10 @@ describeOneTableDetails(const char *schemaname,
 
 	/* Get general table info */
 	printfPQExpBuffer(&buf,
-	 "SELECT relhasindex, relkind, relchecks, reltriggers, relhasrules\n"
+	"SELECT relhasindex, relkind, relchecks, reltriggers, relhasrules, \n"
+					  "relhasoids %s \n"
 					  "FROM pg_catalog.pg_class WHERE oid = '%s'",
+					  pset.sversion >= 80000 ? ", reltablespace" : "",
 					  oid);
 	res = PSQLexec(buf.data, false);
 	if (!res)
@@ -694,11 +731,14 @@ describeOneTableDetails(const char *schemaname,
 	}
 
 	/* FIXME: check for null pointers here? */
-	tableinfo.hasindex = strcmp(PQgetvalue(res, 0, 0), "t") == 0;
-	tableinfo.relkind = *(PQgetvalue(res, 0, 1));
 	tableinfo.checks = atoi(PQgetvalue(res, 0, 2));
 	tableinfo.triggers = atoi(PQgetvalue(res, 0, 3));
+	tableinfo.relkind = *(PQgetvalue(res, 0, 1));
+	tableinfo.hasindex = strcmp(PQgetvalue(res, 0, 0), "t") == 0;
 	tableinfo.hasrules = strcmp(PQgetvalue(res, 0, 4), "t") == 0;
+	tableinfo.hasoids = strcmp(PQgetvalue(res, 0, 5), "t") == 0;
+	tableinfo.tablespace = (pset.sversion >= 80000) ?
+		atooid(PQgetvalue(res, 0, 6)) : 0;
 	PQclear(res);
 
 	headers[0] = _("Column");
@@ -752,14 +792,14 @@ describeOneTableDetails(const char *schemaname,
 			goto error_return;
 
 		if (PQntuples(result) > 0)
-			view_def = xstrdup(PQgetvalue(result, 0, 0));
+			view_def = pg_strdup(PQgetvalue(result, 0, 0));
 
 		PQclear(result);
 	}
 
 	/* Generate table cells to be printed */
 	/* note: initialize all cells[] to NULL in case of error exit */
-	cells = xmalloczero((numrows * cols + 1) * sizeof(*cells));
+	cells = pg_malloc_zero((numrows * cols + 1) * sizeof(*cells));
 
 	for (i = 0; i < numrows; i++)
 	{
@@ -797,9 +837,9 @@ describeOneTableDetails(const char *schemaname,
 			}
 
 #ifdef WIN32
-			cells[i * cols + 2] = xstrdup(mbvalidate(tmpbuf.data, myopt.encoding));
+			cells[i * cols + 2] = pg_strdup(mbvalidate(tmpbuf.data, myopt.encoding));
 #else
-			cells[i * cols + 2] = xstrdup(tmpbuf.data);
+			cells[i * cols + 2] = pg_strdup(tmpbuf.data);
 #endif
 		}
 
@@ -856,8 +896,8 @@ describeOneTableDetails(const char *schemaname,
 		PGresult   *result;
 
 		printfPQExpBuffer(&buf,
-		  "SELECT i.indisunique, i.indisprimary, a.amname, c2.relname,\n"
-					  "  pg_catalog.pg_get_expr(i.indpred, i.indrelid)\n"
+						  "SELECT i.indisunique, i.indisprimary, i.indisclustered, a.amname, c2.relname,\n"
+				"  pg_catalog.pg_get_expr(i.indpred, i.indrelid, true)\n"
 						  "FROM pg_catalog.pg_index i, pg_catalog.pg_class c, pg_catalog.pg_class c2, pg_catalog.pg_am a\n"
 						  "WHERE i.indexrelid = c.oid AND c.oid = '%s' AND c.relam = a.oid\n"
 						  "AND i.indrelid = c2.oid",
@@ -875,9 +915,11 @@ describeOneTableDetails(const char *schemaname,
 		{
 			char	   *indisunique = PQgetvalue(result, 0, 0);
 			char	   *indisprimary = PQgetvalue(result, 0, 1);
-			char	   *indamname = PQgetvalue(result, 0, 2);
-			char	   *indtable = PQgetvalue(result, 0, 3);
-			char	   *indpred = PQgetvalue(result, 0, 4);
+			char	   *indisclustered = PQgetvalue(result, 0, 2);
+			char	   *indamname = PQgetvalue(result, 0, 3);
+			char	   *indtable = PQgetvalue(result, 0, 4);
+			char	   *indpred = PQgetvalue(result, 0, 5);
+			int			count_footers = 0;
 
 			if (strcmp(indisprimary, "t") == 0)
 				printfPQExpBuffer(&tmpbuf, _("primary key, "));
@@ -892,11 +934,17 @@ describeOneTableDetails(const char *schemaname,
 							  schemaname, indtable);
 
 			if (strlen(indpred))
-				appendPQExpBuffer(&tmpbuf, ", predicate %s", indpred);
+				appendPQExpBuffer(&tmpbuf, _(", predicate (%s)"), indpred);
+
+			if (strcmp(indisclustered, "t") == 0)
+				appendPQExpBuffer(&tmpbuf, _(", clustered"));
+
+			footers = pg_malloc_zero(4 * sizeof(*footers));
+			footers[count_footers++] = pg_strdup(tmpbuf.data);
+			add_tablespace_footer(tableinfo.relkind, tableinfo.tablespace,
+								  footers, &count_footers, tmpbuf);
+			footers[count_footers] = NULL;
 
-			footers = xmalloczero(2 * sizeof(*footers));
-			footers[0] = xstrdup(tmpbuf.data);
-			footers[1] = NULL;
 		}
 
 		PQclear(result);
@@ -911,7 +959,7 @@ describeOneTableDetails(const char *schemaname,
 		if (tableinfo.hasrules)
 		{
 			printfPQExpBuffer(&buf,
-							  "SELECT r.rulename\n"
+							  "SELECT r.rulename, trim(trailing ';' from pg_catalog.pg_get_ruledef(r.oid, true))\n"
 							  "FROM pg_catalog.pg_rewrite r\n"
 				   "WHERE r.ev_class = '%s' AND r.rulename != '_RETURN'",
 							  oid);
@@ -923,27 +971,31 @@ describeOneTableDetails(const char *schemaname,
 		}
 
 		/* Footer information about a view */
-		footers = xmalloczero((rule_count + 2) * sizeof(*footers));
-		footers[count_footers] = xmalloc(64 + strlen(view_def));
+		footers = pg_malloc_zero((rule_count + 3) * sizeof(*footers));
+		footers[count_footers] = pg_malloc(64 + strlen(view_def));
 		snprintf(footers[count_footers], 64 + strlen(view_def),
 				 _("View definition:\n%s"), view_def);
 		count_footers++;
 
 		/* print rules */
-		for (i = 0; i < rule_count; i++)
+		if (rule_count > 0)
 		{
-			char	   *s = _("Rules");
+			printfPQExpBuffer(&buf, _("Rules:"));
+			footers[count_footers++] = pg_strdup(buf.data);
+			for (i = 0; i < rule_count; i++)
+			{
+				const char *ruledef;
 
-			if (i == 0)
-				printfPQExpBuffer(&buf, "%s: %s", s, PQgetvalue(result, i, 0));
-			else
-				printfPQExpBuffer(&buf, "%*s  %s", (int) strlen(s), "", PQgetvalue(result, i, 0));
-			if (i < rule_count - 1)
-				appendPQExpBuffer(&buf, ",");
+				/* Everything after "CREATE RULE" is echoed verbatim */
+				ruledef = PQgetvalue(result, i, 1);
+				ruledef += 12;
 
-			footers[count_footers++] = xstrdup(buf.data);
+				printfPQExpBuffer(&buf, " %s", ruledef);
+
+				footers[count_footers++] = pg_strdup(buf.data);
+			}
+			PQclear(result);
 		}
-		PQclear(result);
 
 		footers[count_footers] = NULL;
 
@@ -969,8 +1021,8 @@ describeOneTableDetails(const char *schemaname,
 		if (tableinfo.hasindex)
 		{
 			printfPQExpBuffer(&buf,
-					 "SELECT c2.relname, i.indisprimary, i.indisunique, "
-							  "pg_catalog.pg_get_indexdef(i.indexrelid)\n"
+							  "SELECT c2.relname, i.indisprimary, i.indisunique, i.indisclustered, "
+					"pg_catalog.pg_get_indexdef(i.indexrelid, 0, true)\n"
 							  "FROM pg_catalog.pg_class c, pg_catalog.pg_class c2, pg_catalog.pg_index i\n"
 							  "WHERE c.oid = '%s' AND c.oid = i.indrelid AND i.indexrelid = c2.oid\n"
 							  "ORDER BY i.indisprimary DESC, i.indisunique DESC, c2.relname",
@@ -987,10 +1039,10 @@ describeOneTableDetails(const char *schemaname,
 		{
 			printfPQExpBuffer(&buf,
 							  "SELECT "
-							  "pg_catalog.pg_get_constraintdef(r.oid, true), "
+						 "pg_catalog.pg_get_constraintdef(r.oid, true), "
 							  "conname\n"
 							  "FROM pg_catalog.pg_constraint r\n"
-							  "WHERE r.conrelid = '%s' AND r.contype = 'c'",
+						   "WHERE r.conrelid = '%s' AND r.contype = 'c'",
 							  oid);
 			result2 = PSQLexec(buf.data, false);
 			if (!result2)
@@ -1006,7 +1058,7 @@ describeOneTableDetails(const char *schemaname,
 		if (tableinfo.hasrules)
 		{
 			printfPQExpBuffer(&buf,
-							  "SELECT r.rulename, trim(trailing ';' from pg_catalog.pg_get_ruledef(r.oid))\n"
+							  "SELECT r.rulename, trim(trailing ';' from pg_catalog.pg_get_ruledef(r.oid, true))\n"
 							  "FROM pg_catalog.pg_rewrite r\n"
 							  "WHERE r.ev_class = '%s'",
 							  oid);
@@ -1051,7 +1103,7 @@ describeOneTableDetails(const char *schemaname,
 		{
 			printfPQExpBuffer(&buf,
 							  "SELECT conname,\n"
-					 "  pg_catalog.pg_get_constraintdef(oid) as condef\n"
+			   "  pg_catalog.pg_get_constraintdef(oid, true) as condef\n"
 							  "FROM pg_catalog.pg_constraint r\n"
 						   "WHERE r.conrelid = '%s' AND r.contype = 'f'",
 							  oid);
@@ -1077,14 +1129,14 @@ describeOneTableDetails(const char *schemaname,
 		else
 			inherits_count = PQntuples(result6);
 
-		footers = xmalloczero((index_count + check_count + rule_count + trigger_count + foreignkey_count + inherits_count + 6)
-							  * sizeof(*footers));
+		footers = pg_malloc_zero((index_count + check_count + rule_count + trigger_count + foreignkey_count + inherits_count + 7 + 1)
+								 * sizeof(*footers));
 
 		/* print indexes */
 		if (index_count > 0)
 		{
 			printfPQExpBuffer(&buf, _("Indexes:"));
-			footers[count_footers++] = xstrdup(buf.data);
+			footers[count_footers++] = pg_strdup(buf.data);
 			for (i = 0; i < index_count; i++)
 			{
 				const char *indexdef;
@@ -1097,20 +1149,22 @@ describeOneTableDetails(const char *schemaname,
 				/* Label as primary key or unique (but not both) */
 				appendPQExpBuffer(&buf,
 							  strcmp(PQgetvalue(result1, i, 1), "t") == 0
-								  ? _(" primary key,") :
+								  ? " PRIMARY KEY," :
 							 (strcmp(PQgetvalue(result1, i, 2), "t") == 0
-							  ? _(" unique,")
+							  ? " UNIQUE,"
 							  : ""));
-
 				/* Everything after "USING" is echoed verbatim */
-				indexdef = PQgetvalue(result1, i, 3);
+				indexdef = PQgetvalue(result1, i, 4);
 				usingpos = strstr(indexdef, " USING ");
 				if (usingpos)
 					indexdef = usingpos + 7;
 
 				appendPQExpBuffer(&buf, " %s", indexdef);
 
-				footers[count_footers++] = xstrdup(buf.data);
+				if (strcmp(PQgetvalue(result1, i, 3), "t") == 0)
+					appendPQExpBuffer(&buf, " CLUSTER");
+
+				footers[count_footers++] = pg_strdup(buf.data);
 			}
 		}
 
@@ -1118,14 +1172,14 @@ describeOneTableDetails(const char *schemaname,
 		if (check_count > 0)
 		{
 			printfPQExpBuffer(&buf, _("Check constraints:"));
-			footers[count_footers++] = xstrdup(buf.data);
+			footers[count_footers++] = pg_strdup(buf.data);
 			for (i = 0; i < check_count; i++)
 			{
 				printfPQExpBuffer(&buf, _("    \"%s\" %s"),
 								  PQgetvalue(result2, i, 1),
 								  PQgetvalue(result2, i, 0));
 
-				footers[count_footers++] = xstrdup(buf.data);
+				footers[count_footers++] = pg_strdup(buf.data);
 			}
 		}
 
@@ -1133,14 +1187,14 @@ describeOneTableDetails(const char *schemaname,
 		if (foreignkey_count > 0)
 		{
 			printfPQExpBuffer(&buf, _("Foreign-key constraints:"));
-			footers[count_footers++] = xstrdup(buf.data);
+			footers[count_footers++] = pg_strdup(buf.data);
 			for (i = 0; i < foreignkey_count; i++)
 			{
 				printfPQExpBuffer(&buf, _("    \"%s\" %s"),
 								  PQgetvalue(result5, i, 0),
 								  PQgetvalue(result5, i, 1));
 
-				footers[count_footers++] = xstrdup(buf.data);
+				footers[count_footers++] = pg_strdup(buf.data);
 			}
 		}
 
@@ -1148,7 +1202,7 @@ describeOneTableDetails(const char *schemaname,
 		if (rule_count > 0)
 		{
 			printfPQExpBuffer(&buf, _("Rules:"));
-			footers[count_footers++] = xstrdup(buf.data);
+			footers[count_footers++] = pg_strdup(buf.data);
 			for (i = 0; i < rule_count; i++)
 			{
 				const char *ruledef;
@@ -1159,7 +1213,7 @@ describeOneTableDetails(const char *schemaname,
 
 				printfPQExpBuffer(&buf, "    %s", ruledef);
 
-				footers[count_footers++] = xstrdup(buf.data);
+				footers[count_footers++] = pg_strdup(buf.data);
 			}
 		}
 
@@ -1167,7 +1221,7 @@ describeOneTableDetails(const char *schemaname,
 		if (trigger_count > 0)
 		{
 			printfPQExpBuffer(&buf, _("Triggers:"));
-			footers[count_footers++] = xstrdup(buf.data);
+			footers[count_footers++] = pg_strdup(buf.data);
 			for (i = 0; i < trigger_count; i++)
 			{
 				const char *tgdef;
@@ -1181,7 +1235,7 @@ describeOneTableDetails(const char *schemaname,
 
 				printfPQExpBuffer(&buf, "    %s", tgdef);
 
-				footers[count_footers++] = xstrdup(buf.data);
+				footers[count_footers++] = pg_strdup(buf.data);
 			}
 		}
 
@@ -1197,9 +1251,20 @@ describeOneTableDetails(const char *schemaname,
 			if (i < inherits_count - 1)
 				appendPQExpBuffer(&buf, ",");
 
-			footers[count_footers++] = xstrdup(buf.data);
+			footers[count_footers++] = pg_strdup(buf.data);
 		}
 
+		if (verbose)
+		{
+			char	   *s = _("Has OIDs");
+
+			printfPQExpBuffer(&buf, "%s: %s", s,
+							  (tableinfo.hasoids ? _("yes") : _("no")));
+			footers[count_footers++] = pg_strdup(buf.data);
+		}
+
+		add_tablespace_footer(tableinfo.relkind, tableinfo.tablespace,
+							  footers, &count_footers, buf);
 		/* end of list marker */
 		footers[count_footers] = NULL;
 
@@ -1251,6 +1316,38 @@ error_return:
 }
 
 
+static void
+add_tablespace_footer(char relkind, Oid tablespace, char **footers,
+					  int *count, PQExpBufferData buf)
+{
+	/* relkinds for which we support tablespaces */
+	if (relkind == 'r' || relkind == 'i')
+	{
+		/*
+		 * We ignore the database default tablespace so that users not
+		 * using tablespaces don't need to know about them.
+		 */
+		if (tablespace != 0)
+		{
+			PGresult   *result1 = NULL;
+
+			printfPQExpBuffer(&buf, "SELECT spcname FROM pg_tablespace \n"
+							  "WHERE oid = '%u';", tablespace);
+			result1 = PSQLexec(buf.data, false);
+			if (!result1)
+				return;
+			/* Should always be the case, but.... */
+			if (PQntuples(result1) > 0)
+			{
+				printfPQExpBuffer(&buf, _("Tablespace: \"%s\""),
+								  PQgetvalue(result1, 0, 0));
+				footers[(*count)++] = pg_strdup(buf.data);
+			}
+			PQclear(result1);
+		}
+	}
+}
+
 /*
  * \du
  *
@@ -1272,12 +1369,13 @@ describeUsers(const char *pattern)
 			"       WHEN u.usesuper THEN CAST('%s' AS pg_catalog.text)\n"
 		 "       WHEN u.usecreatedb THEN CAST('%s' AS pg_catalog.text)\n"
 					  "       ELSE CAST('' AS pg_catalog.text)\n"
-					  "  END AS \"%s\"\n"
+					  "  END AS \"%s\",\n"
+					  "  ARRAY(SELECT g.groname FROM pg_catalog.pg_group g WHERE u.usesysid = ANY(g.grolist)) as \"%s\"\n"
 					  "FROM pg_catalog.pg_user u\n",
 					  _("User name"), _("User ID"),
 					  _("superuser, create database"),
 					  _("superuser"), _("create database"),
-					  _("Attributes"));
+					  _("Attributes"), _("Groups"));
 
 	processNamePattern(&buf, pattern, false, false,
 					   NULL, "u.usename", NULL, NULL);
@@ -1290,7 +1388,47 @@ describeUsers(const char *pattern)
 		return false;
 
 	myopt.nullPrint = NULL;
-	myopt.title = _("List of database users");
+	myopt.title = _("List of users");
+
+	printQuery(res, &myopt, pset.queryFout);
+
+	PQclear(res);
+	return true;
+}
+
+
+/*
+ * \dg
+ *
+ * Describes groups.
+ */
+bool
+describeGroups(const char *pattern)
+{
+	PQExpBufferData buf;
+	PGresult   *res;
+	printQueryOpt myopt = pset.popt;
+
+	initPQExpBuffer(&buf);
+
+	printfPQExpBuffer(&buf,
+					  "SELECT g.groname AS \"%s\",\n"
+					  "  g.grosysid AS \"%s\"\n"
+					  "FROM pg_catalog.pg_group g\n",
+					  _("Group name"), _("Group ID"));
+
+	processNamePattern(&buf, pattern, false, false,
+					   NULL, "g.groname", NULL, NULL);
+
+	appendPQExpBuffer(&buf, "ORDER BY 1;");
+
+	res = PSQLexec(buf.data, false);
+	termPQExpBuffer(&buf);
+	if (!res)
+		return false;
+
+	myopt.nullPrint = NULL;
+	myopt.title = _("List of groups");
 
 	printQuery(res, &myopt, pset.queryFout);
 
@@ -1339,6 +1477,11 @@ listTables(const char *tabtypes, const char *pattern, bool verbose)
 					  _("table"), _("view"), _("index"), _("sequence"),
 					  _("special"), _("Type"), _("Owner"));
 
+	if (showIndexes)
+		appendPQExpBuffer(&buf,
+						  ",\n c2.relname as \"%s\"",
+						  _("Table"));
+
 	if (verbose)
 		appendPQExpBuffer(&buf,
 		  ",\n  pg_catalog.obj_description(c.oid, 'pg_class') as \"%s\"",
@@ -1346,13 +1489,11 @@ listTables(const char *tabtypes, const char *pattern, bool verbose)
 
 	if (showIndexes)
 		appendPQExpBuffer(&buf,
-						  ",\n c2.relname as \"%s\""
 						  "\nFROM pg_catalog.pg_class c"
 			  "\n     JOIN pg_catalog.pg_index i ON i.indexrelid = c.oid"
 			  "\n     JOIN pg_catalog.pg_class c2 ON i.indrelid = c2.oid"
 		"\n     LEFT JOIN pg_catalog.pg_user u ON u.usesysid = c.relowner"
-						  "\n     LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace\n",
-						  _("Table"));
+						  "\n     LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace\n");
 	else
 		appendPQExpBuffer(&buf,
 						  "\nFROM pg_catalog.pg_class c"
@@ -1573,7 +1714,7 @@ listCasts(const char *pattern)
  * Describes schemas (namespaces)
  */
 bool
-listSchemas(const char *pattern)
+listSchemas(const char *pattern, bool verbose)
 {
 	PQExpBufferData buf;
 	PGresult   *res;
@@ -1582,13 +1723,22 @@ listSchemas(const char *pattern)
 	initPQExpBuffer(&buf);
 	printfPQExpBuffer(&buf,
 					  "SELECT n.nspname AS \"%s\",\n"
-					  "       u.usename AS \"%s\"\n"
-		"FROM pg_catalog.pg_namespace n LEFT JOIN pg_catalog.pg_user u\n"
-					  "       ON n.nspowner=u.usesysid\n",
-					  _("Name"),
-					  _("Owner"));
+					  "       u.usename AS \"%s\"",
+					  _("Name"), _("Owner"));
 
-	processNamePattern(&buf, pattern, false, false,
+	if (verbose)
+		appendPQExpBuffer(&buf,
+						  ",\n  n.nspacl as \"%s\","
+		 "  pg_catalog.obj_description(n.oid, 'pg_namespace') as \"%s\"",
+						  _("Access privileges"), _("Description"));
+
+	appendPQExpBuffer(&buf,
+	  "\nFROM pg_catalog.pg_namespace n LEFT JOIN pg_catalog.pg_user u\n"
+					  "       ON n.nspowner=u.usesysid\n"
+				 "WHERE	(n.nspname NOT LIKE 'pg\\\\_temp\\\\_%%' OR\n"
+	   "		 n.nspname = (pg_catalog.current_schemas(true))[1])\n");		/* temp schema is first */
+
+	processNamePattern(&buf, pattern, true, false,
 					   NULL, "n.nspname", NULL,
 					   NULL);
 
@@ -1665,34 +1815,37 @@ processNamePattern(PQExpBuffer buf, const char *pattern,
 
 	while (*cp)
 	{
-		if (*cp == '"')
+		char	ch = *cp;
+
+		if (ch == '"')
 		{
 			if (inquotes && cp[1] == '"')
 			{
-				/* emit one quote */
+				/* emit one quote, stay in inquotes mode */
 				appendPQExpBufferChar(&namebuf, '"');
 				cp++;
 			}
-			inquotes = !inquotes;
+			else
+				inquotes = !inquotes;
 			cp++;
 		}
-		else if (!inquotes && isupper((unsigned char) *cp))
+		else if (!inquotes && isupper((unsigned char) ch))
 		{
 			appendPQExpBufferChar(&namebuf,
-								  tolower((unsigned char) *cp));
+								  pg_tolower((unsigned char) ch));
 			cp++;
 		}
-		else if (!inquotes && *cp == '*')
+		else if (!inquotes && ch == '*')
 		{
 			appendPQExpBuffer(&namebuf, ".*");
 			cp++;
 		}
-		else if (!inquotes && *cp == '?')
+		else if (!inquotes && ch == '?')
 		{
 			appendPQExpBufferChar(&namebuf, '.');
 			cp++;
 		}
-		else if (!inquotes && *cp == '.')
+		else if (!inquotes && ch == '.')
 		{
 			/* Found schema/name separator, move current pattern to schema */
 			resetPQExpBuffer(&schemabuf);
@@ -1736,16 +1889,11 @@ processNamePattern(PQExpBuffer buf, const char *pattern,
 	{
 		/* We have a schema pattern, so constrain the schemavar */
 
-		appendPQExpBufferChar(&schemabuf, '$');
-		/* Optimize away ".*$", and possibly the whole pattern */
-		if (schemabuf.len >= 3 &&
-			strcmp(schemabuf.data + (schemabuf.len - 3), ".*$") == 0)
-			schemabuf.data[schemabuf.len - 3] = '\0';
-
-		if (schemabuf.data[0] && schemavar)
+		/* Optimize away a "*" pattern */
+		if (strcmp(schemabuf.data, ".*") != 0 && schemavar)
 		{
 			WHEREAND();
-			appendPQExpBuffer(buf, "%s ~ '^%s'\n",
+			appendPQExpBuffer(buf, "%s ~ '^(%s)$'\n",
 							  schemavar, schemabuf.data);
 		}
 	}
@@ -1763,24 +1911,19 @@ processNamePattern(PQExpBuffer buf, const char *pattern,
 	{
 		/* We have a name pattern, so constrain the namevar(s) */
 
-		appendPQExpBufferChar(&namebuf, '$');
-		/* Optimize away ".*$", and possibly the whole pattern */
-		if (namebuf.len >= 3 &&
-			strcmp(namebuf.data + (namebuf.len - 3), ".*$") == 0)
-			namebuf.data[namebuf.len - 3] = '\0';
-
-		if (namebuf.data[0])
+		/* Optimize away a "*" pattern */
+		if (strcmp(namebuf.data, ".*") != 0)
 		{
 			WHEREAND();
 			if (altnamevar)
 				appendPQExpBuffer(buf,
-								  "(%s ~ '^%s'\n"
-								  "        OR %s ~ '^%s')\n",
+								  "(%s ~ '^(%s)$'\n"
+								  "        OR %s ~ '^(%s)$')\n",
 								  namevar, namebuf.data,
 								  altnamevar, namebuf.data);
 			else
 				appendPQExpBuffer(buf,
-								  "%s ~ '^%s'\n",
+								  "%s ~ '^(%s)$'\n",
 								  namevar, namebuf.data);
 		}
 	}
